part of nueli.core.config;

/// Ink shortcut
Ink get $ink => Ink();

/// Ink ~ [InkWell] & [InkResponse] configuration class.
class Ink extends Config<Ink, InkData> {
  Color? _splash;
  Color? _hover;
  Color? _focus;
  Color? _overlay;
  Color? _highlight;
  Radii? _radii;

  Ink() : super(InkData());
  factory Ink.ripple(Color color, [int alpha = 100]) => Ink().splash(color, alpha ~/ 2).hover(color, alpha).overlay(color, alpha);
  factory Ink.rumble(Color color, [int alpha = 100]) => Ink().highlight(color, alpha).focus(color, alpha).splash(color, alpha);

  Ink _config({
    Color? splash,
    Color? hover,
    Color? focus,
    Color? overlay,
    Color? highlight,
    Radii? radii,
  }) {
    _splash = splash ?? _splash;
    _hover = hover ?? _hover;
    _focus = focus ?? _focus;
    _overlay = overlay ?? _overlay;
    _highlight = highlight ?? _highlight;
    _radii = radii ?? _radii;
    return this;
  }

  // variant commands
  Ink splash(Color color, [int? alpha]) => _config(splash: color.withAlpha(alpha ?? palette.tint.alpha));
  Ink hover(Color color, [int? alpha]) => _config(hover: color.withAlpha(alpha ?? palette.tint.alpha));
  Ink focus(Color color, [int? alpha]) => _config(focus: color.withAlpha(alpha ?? palette.tint.alpha));
  Ink overlay(Color color, [int? alpha]) => _config(overlay: color.withAlpha(alpha ?? palette.tint.alpha));
  Ink highlight(Color color, [int? alpha]) => _config(highlight: color.withAlpha(alpha ?? palette.tint.alpha));

  // fixed commands
  Ink withType(InkType? type) => this & type;
  Ink withTap(Function()? onTap) => this & onTap;
  Ink withLongPress(Function()? onLongPress) => this & Attr(#longpress)[onLongPress];

  // related config reference
  Ink radii$(Radii radii) => _config(radii: radii);

  // templates
  Ink $ripple(Color color, [int alpha = 100]) => Ink.ripple(color, alpha);
  Ink $rumble(Color color, [int alpha = 100]) => Ink.rumble(color, alpha);

  @override
  Ink operator *(double factor) {
    final inkAp = _InkApprox(this);
    return _config(
      splash: inkAp.splash * factor,
      hover: inkAp.hover * factor,
      focus: inkAp.focus * factor,
      overlay: inkAp.overlay * factor,
      highlight: inkAp.highlight * factor,
      radii: inkAp.radii * factor,
    );
  }

  @override
  Ink operator -(Ink operand) {
    final inkAp = _InkApprox(this);
    final inkAp2 = _InkApprox(operand);
    return _config(
      splash: inkAp.splash - inkAp2.splash,
      hover: inkAp.hover - inkAp2.hover,
      focus: inkAp.focus - inkAp2.focus,
      overlay: inkAp.overlay - inkAp2.overlay,
      highlight: inkAp.highlight - inkAp2.highlight,
      radii: inkAp.radii - inkAp2.radii,
    );
  }

  @override
  Ink operator +(Ink operand) {
    final inkAp = _InkApprox(this);
    final inkAp2 = _InkApprox(operand);
    return _config(
      splash: inkAp.splash + inkAp2.splash,
      hover: inkAp.hover + inkAp2.hover,
      focus: inkAp.focus + inkAp2.focus,
      overlay: inkAp.overlay + inkAp2.overlay,
      highlight: inkAp.highlight + inkAp2.highlight,
      radii: inkAp.radii + inkAp2.radii,
    );
  }

  @override
  Ink lerp(ink) {
    if (ink is Ink)
      _config(
        splash: ink._splash ?? _splash,
        hover: ink._hover ?? _hover,
        focus: ink._focus ?? _focus,
        overlay: ink._overlay ?? _overlay,
        highlight: ink._highlight ?? _highlight,
        radii: ink._radii ?? _radii,
      );
    return super.lerp(ink);
  }

  Widget call() {
    final inkAp = _InkApprox(this);

    return data.type == InkType.response
        ? InkResponse(
            child: data.child,
            highlightColor: inkAp.overlay,
            overlayColor: MaterialStateProperty.all<Color>(inkAp.overlay),
            splashColor: inkAp.splash,
            hoverColor: inkAp.hover,
            focusColor: inkAp.focus,
            onTap: this.data.onTap,
            onLongPress: this.data.onLongPress,
          )
        : InkWell(
            child: data.child,
            borderRadius: inkAp.radii.toBorderRadius,
            splashColor: inkAp.splash,
            hoverColor: inkAp.hover,
            focusColor: inkAp.focus,
            overlayColor: MaterialStateProperty.all<Color>(inkAp.overlay),
            highlightColor: inkAp.highlight,
            onTap: data.onTap,
            onLongPress: data.onLongPress,
          );
  }
}

class InkData extends ConfigData<InkData> {
  InkType? type;
  Function()? onTap;
  Function()? onLongPress;
  InkData({Widget? child, this.type, this.onTap, this.onLongPress}) : super(child);

  InkData lerp(ink) {
    if (ink is InkType) type = ink;
    if (ink is Function()) onTap = ink;

    if (ink is Attr && ink is Function()) {
      if (ink.embedded == #tap) onTap = ink.value;
      if (ink.embedded == #longpress) onLongPress = ink.value;
    }

    if (ink is InkData) {
      type = ink.type;
      if (ink is InkData) onTap = ink.onTap;
      if (ink is InkData) onLongPress = ink.onLongPress;
    }
    return super.lerp(ink);
  }
}

class InkTween extends Tween<Ink> {
  InkTween({Ink? begin, Ink? end}) : super(begin: begin, end: end);

  @override
  Ink transform(double value) => (begin ?? $ink) + ((end ?? $ink) - (begin ?? $ink)) * value;
}

class InkTransition extends AnimatedWidget {
  final Widget? child;
  InkTransition({required Animation<Ink> ink, this.child}) : super(listenable: ink);
  Animation<Ink> get ink => listenable as Animation<Ink>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => ink.value.withChild(child)(),
    );
  }
}

enum InkType { response, well }

class _InkApprox {
  final Ink data;
  _InkApprox(this.data);

  Color get splash => data._splash ?? palette.accent.withAlpha(palette.tint.alpha);
  Color get hover => data._hover ?? palette.accent.withAlpha(palette.tint.alpha ~/ 2);
  Color get focus => data._focus ?? palette.accent.withAlpha(palette.tint.alpha ~/ 5);
  Color get overlay => data._overlay ?? palette.accent.withAlpha(palette.tint.alpha ~/ 2);
  Color get highlight => data._highlight ?? palette.accent.withAlpha(palette.tint.alpha ~/ 3);
  Radii get radii => data._radii ?? $radii.$zero;
}
