part of nueli.core.config;

/// Radii shortcut
Radii get $radii => Radii();

/// Radii ~ [BorderRadius] configuration class.
class Radii extends Config<Radii, RadiiData> {
  double? _tlX;
  double? _tlY;
  double? _trX;
  double? _trY;
  double? _brX;
  double? _brY;
  double? _blX;
  double? _blY;

  Radii() : super(RadiiData());
  factory Radii.vert(double top, double bottom) => Radii().tl(top).tr(top).bl(bottom).br(bottom);
  factory Radii.horz(double left, double right) => Radii().tl(left).tr(right).bl(left).br(right);
  factory Radii.corners(double tl, double tr, double br, double bl) => Radii().tl(tl).tr(tr).br(br).bl(bl);
  factory Radii.sides(double left, double right, double top, double bottom) => Radii().left(left).right(right).top(top).bottom(bottom);
  factory Radii.all(double value) => Radii().left(value).right(value).top(value).bottom(value);

  Radii _config({
    double? tlX,
    double? tlY,
    double? trX,
    double? trY,
    double? brX,
    double? brY,
    double? blX,
    double? blY,
  }) {
    _tlX = tlX ?? _tlX;
    _tlY = tlY ?? _tlY;
    _trX = trX ?? _trX;
    _trY = trY ?? _trY;
    _brX = brX ?? _brX;
    _brY = brY ?? _brY;
    _blX = blX ?? _blX;
    _blY = blY ?? _blY;
    return this;
  }

  // variant commands
  Radii tl_x(double value) => _config(tlX: value); // ignore: non_constant_identifier_names
  Radii tl_y(double value) => _config(tlY: value); // ignore: non_constant_identifier_names
  Radii tr_x(double value) => _config(trX: value); // ignore: non_constant_identifier_names
  Radii tr_y(double value) => _config(trY: value); // ignore: non_constant_identifier_names
  Radii br_x(double value) => _config(brX: value); // ignore: non_constant_identifier_names
  Radii br_y(double value) => _config(brY: value); // ignore: non_constant_identifier_names
  Radii bl_x(double value) => _config(blX: value); // ignore: non_constant_identifier_names
  Radii bl_y(double value) => _config(blY: value); // ignore: non_constant_identifier_names

  Radii tl(double x, [double? y]) => tl_x(x).tl_y(y ?? x);
  Radii tr(double x, [double? y]) => tr_x(x).tr_y(y ?? x);
  Radii br(double x, [double? y]) => br_x(x).br_y(y ?? x);
  Radii bl(double x, [double? y]) => bl_x(x).bl_y(y ?? x);
  Radii left(double value) => tl_y(value).bl_y(value);
  Radii right(double value) => tr_y(value).br_y(value);
  Radii top(double value) => tl_x(value).tr_x(value);
  Radii bottom(double value) => bl_x(value).br_x(value);

  // templates
  Radii $vert(double top, double bottom) => Radii.vert(top, bottom);
  Radii $horz(double left, double right) => Radii.horz(left, right);
  Radii $corners(double tl, double tr, double br, double bl) => Radii.corners(tl, tr, br, bl);
  Radii $sides(double left, double right, double top, double bottom) => Radii.sides(left, right, top, bottom);
  Radii $all(double value) => Radii.all(value);
  Radii get $8 => $all(8);
  Radii get $16 => $all(16);
  Radii get $24 => $all(24);
  Radii get $52 => $all(52);
  Radii get $108 => $all(108);
  Radii get $zero => $all(0);

  @override
  Radii operator *(double factor) {
    final radiiAp = _RadiiApprox(this);
    return _config(
      tlX: radiiAp.tlX * factor,
      tlY: radiiAp.tlY * factor,
      trX: radiiAp.trX * factor,
      trY: radiiAp.trY * factor,
      brX: radiiAp.brX * factor,
      brY: radiiAp.brY * factor,
      blX: radiiAp.blX * factor,
      blY: radiiAp.blY * factor,
    );
  }

  @override
  Radii operator -(Radii operand) {
    final radiiAp = _RadiiApprox(this);
    final radiiAp2 = _RadiiApprox(operand);
    return _config(
      tlX: radiiAp.tlX - radiiAp2.tlX,
      tlY: radiiAp.tlY - radiiAp2.tlY,
      trX: radiiAp.trX - radiiAp2.trX,
      trY: radiiAp.trY - radiiAp2.trY,
      brX: radiiAp.brX - radiiAp2.brX,
      brY: radiiAp.brY - radiiAp2.brY,
      blX: radiiAp.blX - radiiAp2.blX,
      blY: radiiAp.blY - radiiAp2.blY,
    );
  }

  @override
  Radii operator +(Radii operand) {
    final radiiAp = _RadiiApprox(this);
    final radiiAp2 = _RadiiApprox(operand);
    return _config(
      tlX: radiiAp.tlX + radiiAp2.tlX,
      tlY: radiiAp.tlY + radiiAp2.tlY,
      trX: radiiAp.trX + radiiAp2.trX,
      trY: radiiAp.trY + radiiAp2.trY,
      brX: radiiAp.brX + radiiAp2.brX,
      brY: radiiAp.brY + radiiAp2.brY,
      blX: radiiAp.blX + radiiAp2.blX,
      blY: radiiAp.blY + radiiAp2.blY,
    );
  }

  @override
  Radii lerp(radii) {
    if (radii is Radii)
      _config(
        tlX: radii._tlX ?? _tlX,
        tlY: radii._tlY ?? _tlY,
        trX: radii._trX ?? _trX,
        trY: radii._trY ?? _trY,
        blX: radii._blX ?? _blX,
        blY: radii._blY ?? _blY,
        brX: radii._brX ?? _brX,
        brY: radii._brY ?? _brY,
      );
    return super.lerp(radii);
  }

  BorderRadius get toBorderRadius {
    final radiiAp = _RadiiApprox(this);
    return BorderRadius.only(
      topLeft: Radius.elliptical(radiiAp.tlX, radiiAp.tlY),
      topRight: Radius.elliptical(radiiAp.trX, radiiAp.trY),
      bottomLeft: Radius.elliptical(radiiAp.blX, radiiAp.blY),
      bottomRight: Radius.elliptical(radiiAp.brX, radiiAp.brY),
    );
  }

  Widget call() {
    return Container(
      child: data.child,
      decoration: BoxDecoration(borderRadius: toBorderRadius),
      clipBehavior: Clip.antiAliasWithSaveLayer,
    );
  }
}

class RadiiData extends ConfigData<RadiiData> {
  RadiiData({Widget? child}) : super(child);
}

class RadiiTween extends Tween<Radii> {
  RadiiTween({Radii? begin, Radii? end}) : super(begin: begin, end: end);

  @override
  Radii transform(double value) => (begin ?? $radii) + ((end ?? $radii) - (begin ?? $radii)) * value;
}

class RadiiTransition extends AnimatedWidget {
  final Widget? child;
  RadiiTransition({required Animation<Radii> radii, this.child}) : super(listenable: radii);
  Animation<Radii> get radii => listenable as Animation<Radii>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => radii.value.withChild(child)(),
    );
  }
}

class _RadiiApprox {
  final Radii radii;
  _RadiiApprox(this.radii);

  double get tlX => radii._tlX ?? 0;
  double get tlY => radii._tlX ?? 0;
  double get trX => radii._trX ?? 0;
  double get trY => radii._trX ?? 0;
  double get blX => radii._blX ?? 0;
  double get blY => radii._blX ?? 0;
  double get brX => radii._brX ?? 0;
  double get brY => radii._brX ?? 0;
}
