part of nueli.core.config;

/// Txt shortcut
Txt get $txt => Txt();

/// Txt ~ [Text] and [TextStyle] configuration class.
class Txt extends Config<Txt, TxtData> {
  List<Txt> _children = [];
  Color? _color, _backgroundColor;
  double? _size;
  Ace? _shadow;
  Ink? _ink;

  Txt() : super(TxtData());

  Txt _config({
    Color? color,
    Color? background,
    double? size,
    Ace? shadow,
    Ink? ink,
  }) {
    _color = color ?? _color;
    _backgroundColor = background ?? _backgroundColor;
    _size = size ?? _size;
    _shadow = shadow ?? _shadow;
    _ink = ink ?? _ink;
    return this;
  }

  // variant commands
  Txt color(Color color) => _config(color: color);
  Txt size(double size) => _config(size: size);

  // fixed commands
  Txt withText(String? text) => this & text;
  Txt withWeight(FontWeight? weight) => this & weight;
  Txt withDecoration(TextDecoration? decoration) => this & decoration;
  Txt withBaseline(TextBaseline? baseline) => this & baseline;
  Txt withOverflow(TextOverflow? overflow) => this & overflow;
  Txt withStyle(FontStyle? style) => this & style;
  Txt withAlign(TextAlign? align) => this & align;

  // shortcuts
  Txt get color_onBackground => color(palette.onBackground); // ignore: non_constant_identifier_names
  Txt get color_hint => color(palette.onBackground.withAlpha(150)); // ignore: non_constant_identifier_names

  Txt get size_10 => size(10);
  Txt get size_12 => size(12);
  Txt get size_14 => size(14);
  Txt get size_16 => size(16);
  Txt get size_18 => size(18);
  Txt get size_24 => size(24);

  // related config reference
  Txt ace$(Ace shadow) => _config(shadow: shadow);
  Txt ink$(Ink react) => _config(ink: react);

  @override
  Txt operator *(double factor) {
    final txtAp = _TxtApprox(this);
    return _config(
      color: txtAp.color * factor,
      background: txtAp.backgroundColor * factor,
      size: txtAp.size * factor,
      shadow: txtAp.shadow * factor,
      ink: txtAp.ink * factor,
    );
  }

  @override
  Txt operator -(Txt operand) {
    final txtAp = _TxtApprox(this);
    final txtAp2 = _TxtApprox(operand);
    return _config(
      color: txtAp.color - txtAp2.color,
      background: txtAp.backgroundColor - txtAp2.backgroundColor,
      size: txtAp.size - txtAp2.size,
      shadow: txtAp.shadow - txtAp2.shadow,
      ink: txtAp.ink - txtAp2.ink,
    );
  }

  @override
  Txt operator +(Txt operand) {
    final txtAp = _TxtApprox(this);
    final txtAp2 = _TxtApprox(operand);
    return _config(
      color: txtAp.color + txtAp2.color,
      background: txtAp.backgroundColor + txtAp2.backgroundColor,
      size: txtAp.size + txtAp2.size,
      shadow: txtAp.shadow + txtAp2.shadow,
      ink: txtAp.ink + txtAp2.ink,
    );
  }

  Txt lerp(txt) {
    if (txt is Txt)
      _config(
        color: txt._color ?? _color,
        background: txt._backgroundColor ?? _backgroundColor,
        size: txt._size ?? _size,
        shadow: txt._shadow ?? _shadow,
        ink: txt._ink ?? _ink,
      );
    return super.lerp(txt);
  }

  Txt save() => Txt().._children.addAll([...this._children, this.._children = []]);

  Txt inline(List<Txt> children) => this.._children = children;

  Txt highlight(String charseq, Txt config) {
    List<String> children = data.text.split(RegExp(charseq));
    List<Match> matches = RegExp(charseq).allMatches(data.text).toList();
    List<Txt> configList = [];
    for (int i = 1; i < children.length; i++) {
      String match = matches[i - 1].group(0) ?? '';
      configList.add(config & match);
      configList.add(this & children[i]);
    }
    return (this & children.first).inline(configList);
  }

  String toString() => data.text;

  TextStyle get toTextStyle {
    final txtAp = _TxtApprox(this);
    return TextStyle(
      color: txtAp.color,
      decoration: TextDecoration.combine(data.decorations),
      fontSize: txtAp.size,
      shadows: txtAp.shadow.toBoxShadowDeep,
      fontStyle: data.style,
      fontWeight: data.weight,
      backgroundColor: txtAp.backgroundColor,
      textBaseline: TextBaseline.alphabetic,
    );
  }

  GestureRecognizer? get toRecognizer {
    final txtAp = _TxtApprox(this);
    if (txtAp.ink.data.onTap != null) return TapGestureRecognizer()..onTap = () => (txtAp.ink.data.onTap ?? () {})();
    if (txtAp.ink.data.onLongPress != null) return LongPressGestureRecognizer()..onLongPress = () => (txtAp.ink.data.onLongPress ?? () {})();
    return null;
  }

  Widget call() {
    final txtAp = _TxtApprox(this);

    if (_children.isNotEmpty) {
      List<TextSpan> spans = [];
      for (Txt config in _children)
        spans.add(TextSpan(
          text: (this | config).data.text,
          style: (this | config).toTextStyle,
          recognizer: (this | config).toRecognizer,
        ));
      return Container(
        width: data.fill ? double.infinity : null,
        child: Text.rich(
            TextSpan(
              text: data.text,
              style: toTextStyle,
              children: spans,
              recognizer: toRecognizer,
            ), // break
            textAlign: data.align,
            maxLines: data.maxLines,
            overflow: data.overflow,
            style: toTextStyle),
      );
    }

    final text = Container(
      width: data.fill ? double.infinity : null,
      child: Text(data.text, // break
          textAlign: data.align,
          maxLines: data.maxLines,
          overflow: data.overflow,
          style: toTextStyle),
    );
    if (_ink == null) return text;
    return (txtAp.ink & text)();
  }
}

class TxtData extends ConfigData<TxtData> {
  bool fill;
  String text;
  int? maxLines;
  FontWeight weight;
  List<TextDecoration> decorations;
  TextBaseline? baseline;
  TextOverflow? overflow;
  FontStyle? style;
  TextAlign align;

  TxtData({
    this.fill = false,
    this.text = '',
    this.maxLines,
    this.weight = FontWeight.normal,
    this.decorations = const [],
    this.baseline,
    this.overflow,
    this.style,
    this.align = TextAlign.start,
  }) : super(SizedBox());

  TxtData lerp(data) {
    if (data is String) text = data;
    if (data is FontWeight) weight = data;
    if (data is List<TextDecoration>) decorations = data;
    if (data is TextDecoration) decorations.add(data);
    if (data is TextBaseline) baseline = data;
    if (data is TextOverflow) overflow = data;
    if (data is FontStyle) style = data;
    if (data is TextAlign) align = data;

    if (data is TxtData) text = data.text;
    if (data is TxtData) maxLines = data.maxLines;
    if (data is TxtData) weight = data.weight;
    if (data is TxtData) decorations = data.decorations;
    if (data is TxtData) baseline = data.baseline;
    if (data is TxtData) overflow = data.overflow;
    if (data is TxtData) style = data.style;
    if (data is TxtData) align = data.align;
    return super.lerp(data);
  }
}

class TxtInk extends InkData {
  final Function(Txt text)? onTextTap;
  final Function(Txt text)? onTextLongPress;
  TxtInk({
    Function()? onTap,
    Function()? onLongPress,
    this.onTextTap,
    this.onTextLongPress,
  }) : super(type: InkType.well, onTap: onTap, onLongPress: onLongPress);
}

class TxtTween extends Tween<Txt> {
  TxtTween({Txt? begin, Txt? end}) : super(begin: begin, end: end);

  @override
  Txt transform(double value) => (begin ?? $txt) + ((end ?? $txt) - (begin ?? $txt)) * value;
}

class TxtTransition extends AnimatedWidget {
  final Widget? child;
  TxtTransition({required Animation<Txt> txt, this.child}) : super(listenable: txt);
  Animation<Txt> get txt => listenable as Animation<Txt>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => txt.value.withChild(child)(),
    );
  }
}

class _TxtApprox {
  final Txt txt;
  _TxtApprox(this.txt);

  Color get color => txt._color ?? palette.onBackground;
  Color get backgroundColor => txt._backgroundColor ?? palette.none;
  double get size => txt._size ?? 12;
  Ace get shadow => txt._shadow ?? $ace;
  Ink get ink => txt._ink ?? $ink;
}
