library nueli.ui;

import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:need_resume/need_resume.dart';

import '../core/--core.dart';

part 'screen.dart';
part 'struct.dart';

const ui = const _UITools();

mixin UI on StatefulWidget {
  final Ref<Function(Function() fn)> _setState = $((fn) => fn());

  @mustCallSuper
  BuildContext get context => ui.globalContext!;

  @mustCallSuper
  Widget get empty => ui.empty;

  @mustCallSuper
  Widget gap(double size) => ui.gap(size);

  @mustCallSuper
  Widget dot(double size, [Color? color]) => ui.dot(size, color ?? palette.onBackground);

  @mustCallSuper
  Widget lineH(double thickness, [Color? color]) => ui.lineH(thickness, color ?? palette.onBackground);

  @mustCallSuper
  Widget lineV(double thickness, [Color? color]) => ui.lineV(thickness, color ?? palette.onBackground);

  @mustCallSuper
  Expanded flex(int flex) => ui.flex(flex);

  @mustCallSuper
  void setState(Function() fn) => _setState.data!(fn);
}

mixin UIState<T extends UI> on State<T> {
  @mustCallSuper
  Widget get empty => ui.empty;

  @mustCallSuper
  Widget gap(double size) => _UITools().gap(size);

  @mustCallSuper
  Widget dot(double size, [Color? color]) => _UITools().dot(size, color ?? palette.onBackground);

  @mustCallSuper
  Widget lineH(double thickness, [Color? color]) => _UITools().lineH(thickness, color ?? palette.onBackground);

  @mustCallSuper
  Widget lineV(double thickness, [Color? color]) => _UITools().lineV(thickness, color ?? palette.onBackground);

  @mustCallSuper
  Expanded flex(int flex) => _UITools().flex(flex);

  @mustCallSuper
  initTools() {
    widget._setState.data = setState;
    return ui.globalContext = context;
  }

  @override
  void initState() {
    initTools();
    super.initState();
  }

  @mustCallSuper
  Widget build(BuildContext context) {
    initTools();
    return empty;
  }
}

class _UITools {
  static BuildContext? _globalContext;
  static const Widget _empty = SizedBox();

  const _UITools();

  set globalContext(BuildContext? context) => _UITools._globalContext = context;

  BuildContext? get globalContext => _globalContext;

  Widget get empty => _empty;

  Widget gap(double size) => SizedBox(width: size, height: size);

  Widget lineH(double thickness, [Color? color]) => Container(
        height: thickness,
        width: double.infinity,
        decoration: $decor.paint(color ?? palette.accent).radii$($radii.$all(thickness)).toDecoration,
      );

  Widget lineV(double thickness, [Color? color]) => Container(
        width: thickness,
        height: double.infinity,
        decoration: $decor.paint(color ?? palette.accent).radii$($radii.$all(thickness)).toDecoration,
      );

  Widget dot(double size, [Color? color]) => Container(
        width: size,
        height: size,
        decoration: $decor.paint(color ?? palette.accent).radii$($radii.$all(size)).toDecoration,
      );

  Expanded flex(int flex) => Expanded(child: empty, flex: flex);
}

extension ExpandedUI on Expanded {
  Expanded operator &(Widget child) => Expanded(child: child, key: key, flex: flex);
}
